class LiveParticleViewer extends SqRootScript
{
	function colPropToArray(colourProperty) //this is used for setting and displaying Bitmap Disk Colour info
	{
		local colourParts = split(colourProperty, ","); //Property is an array of useless strings. Down with strings!
		local newColourArray = array(3,0); //Creates an array of beautiful integers
		
		newColourArray[0] = colourParts[0].tointeger();
		newColourArray[1] = colourParts[1].tointeger();
		newColourArray[2] = colourParts[2].tointeger();
		
		return newColourArray;
	}
	
	function copy3Props(sourceObj, destObj, alternateStorageOfActiveBool)
	{
		Property.CopyFrom(destObj, "ParticleGroup", sourceObj);
		Property.CopyFrom(destObj, "PGLaunchInfo", sourceObj);
		Property.CopyFrom(destObj, "RenderAlpha", sourceObj);
		
		//undo or redo objects may need the active setting to be disabled in case objects expire.
		//actual value to be stored in otherwise fake property of the same type
		if(alternateStorageOfActiveBool)
		{
			//store in fake property for retrieval later
			local activePropValue = Property.Get(destObj, "ParticleGroup", "Active");
			Property.SetSimple(destObj, "Immobile", activePropValue);
			Property.Set(destObj, "ParticleGroup", "Active", 0); //make false to prevent expiry
		}
		else
		{
			//retrieve from fake property
			local fakeActivePropValue = Property.Get(sourceObj, "Immobile");
			Property.Set(destObj, "ParticleGroup", "Active", fakeActivePropValue);
		}
	}
	
	function getDestOfSPLinkWithData(fromObjID, dataToCheck, deleteLinkAfterFindingObjID = false)
	{
		local linkDestObjID = 0;
		local linkID = 0;
		if(Link.AnyExist("ScriptParams", fromObjID))
		{
			local spLinks = Link.GetAll("ScriptParams", fromObjID);
			foreach(spL in spLinks)
			{
				local linkData = LinkTools.LinkGetData(spL, "");
				if(linkData == dataToCheck)
				{
					local linkInfo = sLink(spL);
					linkDestObjID = linkInfo.dest;
					linkID = spL;
				}
			}
		}
		if(deleteLinkAfterFindingObjID && linkID != 0) Link.Destroy(linkID);
		return linkDestObjID;
	}
}

class ShowParticleInfo extends LiveParticleViewer
{
	function OnSim()
	{
		if(message().starting)
		{
			DisplayParticlesPropertyValues();
		}
	}
	
	function OnTurnOn()
	{
		DisplayParticlesPropertyValues();
	}
	
	function DisplayParticlesPropertyValues() //this also gets called at the end of the OnFrobWorldEnd() function
	{
		local sfxObj = "SFXTest";
		local particleProp = "ParticleGroup";
		local launchProp = "PGLaunchInfo";
		//Set QVars from existing particle property
		//and display property values on object in game
		
		//Particles property
		showCheckBoxValue(sfxObj, particleProp, "Active", "Active");
		
		local partRendType = Property.Get(sfxObj, particleProp, "Particle Render Type");
		Property.SetSimple("PartRendType", "otxtrepr0", "obj\\txt16\\PartType" + partRendType);
		
		local partAnim = Property.Get(sfxObj, particleProp, "Particle Animation");
		Property.SetSimple("PartAnim", "otxtrepr0", "obj\\txt16\\PartAnim" + partAnim);
		local partGroupMotion = Property.Get(sfxObj, particleProp, "Particle Group Motion");
		Property.SetSimple("PartGroupMot", "otxtrepr0", "obj\\txt16\\PtGroupMot" + partGroupMotion);
		
		showPosInteger("numparts", sfxObj, particleProp, "number of particles");
		showDecimalValue("sizepart", sfxObj, particleProp, "size of particle");
		
		showBitmapName(Property.Get(sfxObj, particleProp, "bitmap name"));
		
		local velocity = Property.Get(sfxObj, particleProp, "velocity");
		showVectorPart(velocity.x, "xVelocity", "xVelocityNegative");
		showVectorPart(velocity.y, "yVelocity", "yVelocityNegative");
		showVectorPart(velocity.z, "zVelocity", "zVelocityNegative");
		
		local gravityVector = Property.Get(sfxObj, particleProp, "gravity vector");
		showVectorPart(gravityVector.x, "xGravVector", "xGravNegative");
		showVectorPart(gravityVector.y, "yGravVector", "yGravNegative");
		showVectorPart(gravityVector.z, "zGravVector", "zGravNegative");
		
		Quest.Set("colour1", Property.Get(sfxObj, particleProp, "color (palettized)"));
		Quest.Set("colour2", Property.Get(sfxObj, particleProp, "2nd color"));
		Quest.Set("colour3", Property.Get(sfxObj, particleProp, "3rd color"));
		Quest.Set("alpha", Property.Get(sfxObj, particleProp, "alpha"));
		
		showDecimalValue("fxGroupRad", sfxObj, particleProp, "fixed-group radius");
		
		showCheckBoxValue(sfxObj, particleProp, "spin / pulse group", "SpinPulseGroup");
		
		local spinSpeed = Property.Get(sfxObj, particleProp, "spin speed");
		showVectorPart(spinSpeed.x, "xSpinSpeed", "xSpinNegative");
		showVectorPart(spinSpeed.y, "ySpinSpeed", "ySpinNegative");
		showVectorPart(spinSpeed.z, "zSpinSpeed", "zSpinNegative");
		
		showDecimalValue("pulseMag", sfxObj, particleProp, "pulse magnitude");
		showPosInteger("pulseCycleTime", sfxObj, particleProp, "pulse cycle time ms");
		showSFloatValue("partFadeTime", sfxObj, particleProp, "particle fade time");
		showSFloatValue("launchPeriod", sfxObj, particleProp, "launch period");
		showPosInteger("AnimOffset", sfxObj, particleProp, "animation offset ms");
		showDecimalValue("GroupScaleVel", sfxObj, particleProp, "Group-scale velocity");
		
		showBMDiskFlags(sfxObj);
		
		showSFloatValue("BmBirthTime", sfxObj, particleProp, "bm-disk birth time");
		
		local bmDiskRot = Property.Get(sfxObj, particleProp, "bm-disk rot (vel, offs, rand)");
		showVectorPart(bmDiskRot.x, "xBmDiskRot", "xBmRotNegative");
		showVectorPart(bmDiskRot.y, "yBmDiskRot", "yBmRotNegative");
		showVectorPart(bmDiskRot.z, "zBmDiskRot", "zBmRotNegative");
		
		showSFloatValue("BmAniFrameTime", sfxObj, particleProp, "bm-disk ani frame time");
		showDecimalValue("BmGrowSpeed", sfxObj, particleProp, "bm-disk grow speed");
		
		showBMDiskColour(sfxObj, "bm-disk rgb", "BmCol1Red", "BmCol1Green", "BmCol1Blue", "bmColPreview1");
		showBMDiskColour(sfxObj, "bm-disk 2nd rgb", "BmCol2Red", "BmCol2Green", "BmCol2Blue", "bmColPreview2");
		showBMDiskColour(sfxObj, "bm-disk 3rd rgb", "BmCol3Red", "BmCol3Green", "BmCol3Blue", "bmColPreview3");
		
		showCheckBoxValue(sfxObj, particleProp, "always simulate (slow)", "AlwaysSimulate");
		showCheckBoxValue(sfxObj, particleProp, "always simulate group", "AlwaysSimulateGroup");
		showCheckBoxValue(sfxObj, particleProp, "sort particles (slow)", "SortParticles"); //automatically set when particle is 'scaled bitmap'
		showCheckBoxValue(sfxObj, particleProp, "particles start launched", "ParticlesStartLaunched");
		showCheckBoxValue(sfxObj, particleProp, "alpha subpixel particles", "AlphaSubpixelParticles");
		showCheckBoxValue(sfxObj, particleProp, "skip subpixel particles", "SkipSubPixelParticles");
		showCheckBoxValue(sfxObj, particleProp, "ignore attachment refs", "IgnoreAttachmentRefs");
		showCheckBoxValue(sfxObj, particleProp, "force matching unrefs", "ForceMatchingUnrefs");
		
		//Launch Info property
		local partLaunchType = Property.Get(sfxObj, launchProp, "Launch Type");
		Property.SetSimple("PartLaunchType", "otxtrepr0", "obj\\txt16\\LaunchType" + partLaunchType);
		
		local boxMin = Property.Get(sfxObj, launchProp, "Box Min");
		showVectorPart(boxMin.x, "xBoxMin", "xBoxMinNegative");
		showVectorPart(boxMin.y, "yBoxMin", "yBoxMinNegative");
		showVectorPart(boxMin.z, "zBoxMin", "zBoxMinNegative");
		
		local boxMax = Property.Get(sfxObj, launchProp, "Box Max");
		showVectorPart(boxMax.x, "xBoxMax", "xBoxMaxNegative");
		showVectorPart(boxMax.y, "yBoxMax", "yBoxMaxNegative");
		showVectorPart(boxMax.z, "zBoxMax", "zBoxMaxNegative");
		
		showDecimalValue("MinRadius", sfxObj, launchProp, "Min Radius");
		showDecimalValue("MaxRadius", sfxObj, launchProp, "Max Radius");
		
		showCheckBoxValue(sfxObj, launchProp, "Loc unrotated?", "LocUnrotated");
		
		local velMin = Property.Get(sfxObj, launchProp, "Velocity Min");
		showVectorPart(velMin.x, "xVelocityMin", "xMinVelNegative");
		showVectorPart(velMin.y, "yVelocityMin", "yMinVelNegative");
		showVectorPart(velMin.z, "zVelocityMin", "zMinVelNegative");
		
		local velMax = Property.Get(sfxObj, launchProp, "Velocity Max");
		showVectorPart(velMax.x, "xVelocityMax", "xMaxVelNegative");
		showVectorPart(velMax.y, "yVelocityMax", "yMaxVelNegative");
		showVectorPart(velMax.z, "zVelocityMax", "zMaxVelNegative");
		
		showCheckBoxValue(sfxObj, launchProp, "Vel unrotated?", "VelUnrotated");
		
		showDecimalValue("MinTime", sfxObj, launchProp, "Min Time");
		showDecimalValue("MaxTime", sfxObj, launchProp, "Max Time");
		
		//Object alpha
		Quest.Set("RenderAlpha", Property.Get(sfxObj, "RenderAlpha")*100);
	}
	
	function showBMDiskColour(sfxObj, subPropName, redQVar, greenQVar, blueQVar, particleFXPreviewObj)
	{
		local colourProp = Property.Get(sfxObj, "ParticleGroup", subPropName);
		Property.Set(particleFXPreviewObj, "ParticleGroup", "bm-disk rgb", colourProp);
		
		local colourArray = colPropToArray(colourProp);
		
		Quest.Set(redQVar, colourArray[0]);
		Quest.Set(greenQVar, colourArray[1]);
		Quest.Set(blueQVar, colourArray[2]);
	}
	
	function showBMDiskFlags(sfxObj)
	{
		local flagsTotal = Property.Get(sfxObj, "ParticleGroup", "bm-disk flags");
		local FLAG_MAX = 8;
		for(local i = 0; i < FLAG_MAX; i++)
		{
			local flagIntValue = 1 << i; //just for you skol
			local flagTex = "obj\\txt16\\BMFlags";
			if(flagsTotal & flagIntValue) flagTex += "HL";
			
			Property.SetSimple("bFlag" + flagIntValue, "otxtrepr0", flagTex);
		}
	}
	
	function showCheckBoxValue(objName, particleProp, subPropName, checkBoxObject)
	{
		local checkBoxValue = Property.Get(objName, particleProp, subPropName);
		Property.SetSimple(checkBoxObject, "otxtrepr0", "obj\\txt16\\check" + checkBoxValue); 
	}
	
	function showPosInteger(qvar, obName, propName, subProp)
	{
		Quest.Set(qvar, Property.Get(obName, propName, subProp));
	}
	
	function showDecimalValue(qvar, obName, propName, subProp)
	{
		local prop = Property.Get(obName, propName, subProp);
		local rounded = round(prop, 2);
		Quest.Set(qvar, rounded * 100);
	}
	
	function showSFloatValue(qVar, obName, propName, subProp)
	{
		local propValue = Property.Get(obName, propName, subProp);
		local floatValue = fromSFloat(propValue);
		local rounded = round(floatValue, 2);
		Quest.Set(qVar, rounded * 100);
	}
	
	function fromSFloat(value)
	{
		return value.tofloat()/65536;
	}
	
	function toSFloat(value)
	{
		return value.tofloat()*65536;
	}
	
	function round(val, decimalPoints) //in this class this on only dealing with positive numbers. Additional check needed if using with negative numbers.
	{
		local f = pow(10, decimalPoints) * 1.0; // * 1.0 ensures the result is a float
		local newVal = val * f;
		newVal = floor(newVal + 0.5)
		newVal = (newVal * 1.0) / f;
		return newVal;
	}
	
	function showBitmapName(bmpName)
	{
		local BITMAP_NAME_LIMIT = 15;
		
		//reset
		for(local i = 0; i < BITMAP_NAME_LIMIT; i+=1) Property.SetSimple("bChar" + i, "otxtrepr0", "obj\\txt16\\charBlank");
		
		//fill with letters
		for(local i = 0; i < bmpName.len(); i+=1) Property.SetSimple("bChar" + i, "otxtrepr0", "obj\\txt16\\char" + bmpName[i].tochar());
	}
	
	function showVectorPart(vectComponent, qVarName, negativeObjectName)
	{
		local rounded = round(vectComponent, 2);
		Quest.Set(qVarName, fabs(rounded)*100);
		if(vectComponent < 0) Property.SetSimple(negativeObjectName, "RenderType", 2); //unlit
		else Property.SetSimple(negativeObjectName, "RenderType", 1); //unrendered
	}
}

class UndoParticles extends LiveParticleViewer
{
	function OnWorldSelect()
	{
		if("Undo" in userparams())
		{
			DarkUI.TextMessage("Undo (all 3 properties)", 0, 25000);
		}
		else if("Redo" in userparams())
		{
			DarkUI.TextMessage("Redo (all 3 properties)", 0, 25000);
		}
	}
	
	function OnWorldDeselect()
	{
		DarkUI.TextMessage("");
	}
	
	function OnFrobWorldEnd()
	{
		if("Undo" in userparams())
		{
			undo();
		}
		else if("Redo" in userparams())
		{
			redo();
		}
		
		Sound.PlaySchemaAtObject(self, "Button_Number", self);
	}
	
	function undo()
	{			
		if(Link.AnyExist("ScriptParams", "SFXTest"))
		{
			local currentUndoObjID = getDestOfSPLinkWithData("SFXTest", "Undo");
	
			//get dest of SP link to previous Undo object
			if(currentUndoObjID != 0)
			{
				local currentRedoObjID = getDestOfSPLinkWithData("SFXTest", "Redo", true);
				//before copying from undo object to SFX test
				//first create a new Redo object
				local redoMarkerID = Object.Create("Marker");
				//then copy the 3 props from SFX test to this
				copy3Props("SFXTest", redoMarkerID, true);
				
				//get location of anchor point
				local anchorLoc = Property.Get("UndoRedoAnchor", "Position", "Location");
				local yPosition = anchorLoc.y + 0.1; //small offset, default location for first Redo object
				
				if(currentRedoObjID != 0)
				{
					//sets its position to the previous Redo object, y + 0.5
					local oldPosition = Property.Get(currentRedoObjID, "Position", "Location");
					yPosition = oldPosition.y + 0.1;
					
					//Create a link from new redo object to old redo object.
					local newSPchainLinkID = Link.Create("ScriptParams", redoMarkerID, currentRedoObjID);
					LinkTools.LinkSetData(newSPchainLinkID, "", "Previous");
				}
				
				Object.Teleport(redoMarkerID, vector(anchorLoc.x, yPosition, anchorLoc.z), vector(0,0,0));
				//Property.SetSimple(redoMarkerID, "RenderType", 2); //Unlit, for testing only
				
				//create a new Redo SP link from SFXTest to the new Redo object
				local newRedoLinkID = Link.Create("ScriptParams", "SFXTest", redoMarkerID);
				LinkTools.LinkSetData(newRedoLinkID, "", "Redo");

				Quest.Set("RedoObjID", redoMarkerID);
				
				//copy props to SFXTest
				copy3Props(currentUndoObjID, "SFXTest", false);
				SendMessage("ParticlesPanel", "TurnOn"); //shows the current values of the particles property
				
				local prevUndoObjID = getDestOfSPLinkWithData(currentUndoObjID, "Previous");
				
				if(prevUndoObjID != 0)
				{
					local newLink = Link.Create("ScriptParams", "SFXTest", prevUndoObjID);
					LinkTools.LinkSetData(newLink, "", "Undo");
					Quest.Set("UndoObjID", prevUndoObjID);
				}
				
				Object.Destroy(currentUndoObjID);				
			}
		}
	}
	
	function redo()
	{
		local currentRedoObjID = getDestOfSPLinkWithData("SFXTest", "Redo");
		
		//get dest of SP link to previous Undo object
		if(currentRedoObjID != 0)
		{	
			local currentUndoObjID = getDestOfSPLinkWithData("SFXTest", "Undo", true);
			local undoMarkerID = Object.Create("Marker");
			copy3Props("SFXTest", undoMarkerID, true);
			
			local anchorLoc = Property.Get("UndoRedoAnchor", "Position", "Location");
			local xPosition = anchorLoc.x + 0.1; //small offset, default location for first Undo object
			
			if(currentUndoObjID != 0)
			{
				//sets its position to the previous Undo object, x + 0.5
				local oldPosition = Property.Get(currentUndoObjID, "Position", "Location");
				xPosition = oldPosition.x + 0.1;
				
				//Create a link from new redo object to old redo object.
				local newSPchainLinkID = Link.Create("ScriptParams", undoMarkerID, currentUndoObjID);
				LinkTools.LinkSetData(newSPchainLinkID, "", "Previous");
			}
			
			Object.Teleport(undoMarkerID, vector(xPosition, anchorLoc.y, anchorLoc.z), vector(0,0,0));
			//Property.SetSimple(undoMarkerID, "RenderType", 2); //Unlit, for testing only
			
			local newUndoLinkID = Link.Create("ScriptParams", "SFXTest", undoMarkerID);
			LinkTools.LinkSetData(newUndoLinkID, "", "Undo");
			
			Quest.Set("UndoObjID", undoMarkerID);
			
			copy3Props(currentRedoObjID, "SFXTest", false);
			SendMessage("ParticlesPanel", "TurnOn"); //shows the current values of the particles property
			
			local prevRedoObjID = getDestOfSPLinkWithData(currentRedoObjID, "Previous");
			
			if(prevRedoObjID != 0)
			{
				local newLink = Link.Create("ScriptParams", "SFXTest", prevRedoObjID);
				LinkTools.LinkSetData(newLink, "", "Redo");
				Quest.Set("RedoObjID", prevRedoObjID);
			}
			
			Object.Destroy(currentRedoObjID);
		}		
	}
}

class ChangeParticles extends LiveParticleViewer
{	
	function OnWorldSelect()
	{
		if("Warning" in userparams())
		{
			DarkUI.TextMessage(userparams().Warning, 0, 25000);
		}
		if("HideValue" in userparams()) return;
		if("ChangeValue" in userparams()) DarkUI.TextMessage(userparams().ChangeValue);
		else if("DefaultSubProp" in userparams()) DarkUI.TextMessage("Reset \"" + userparams().DefaultSubProp + "\" to default value from archetype.");
		else if("DefaultPropSimple" in userparams()) DarkUI.TextMessage("Reset \"" + userparams().DefaultPropSimple + "\" to default value from archetype.");
	}
	function OnWorldDeselect()
	{
		DarkUI.TextMessage("");
	}
	
	function updateUndoRedoStack()
	{	
		//Delete any existing SP link to an undo object, ready to create a new one
		local oldUndoMarkerID = getDestOfSPLinkWithData("SFXTest", "Undo", true);
		
		//get location of anchor point
		local anchorLoc = Property.Get("UndoRedoAnchor", "Position", "Location");
		local xPosition = anchorLoc.x + 0.5; //small offset
		
		local undoMarkerID = Object.Create("Marker");
		
		//Create new SP link to undo object
		local spID = Link.Create("ScriptParams", "SFXTest", undoMarkerID);
		LinkTools.LinkSetData(spID, "", "Undo");
		Quest.Set("UndoObjID", undoMarkerID);
		
		//link new undo object to old undo object
		if(oldUndoMarkerID != 0)
		{
			local newSPLinkID = Link.Create("ScriptParams", undoMarkerID, oldUndoMarkerID);
			LinkTools.LinkSetData(newSPLinkID, "", "Previous");
			local oldMarkerPos = Property.Get(oldUndoMarkerID, "Position", "Location");
			local oldMarkerX = oldMarkerPos.x;
			xPosition = oldMarkerX + 0.5;
		}
		
		Object.Teleport(undoMarkerID, vector(xPosition, anchorLoc.y, anchorLoc.z), vector(0,0,0));
		//Property.SetSimple(undoMarkerID, "RenderType", 2); //Unlit, for testing only
		
		//copy current 3 props to it
		copy3Props("SFXTest", undoMarkerID, true);

		local redoStackStartObjID = getDestOfSPLinkWithData("SFXTest", "Redo");
		if(redoStackStartObjID) deleteRedoStack(redoStackStartObjID);
	}
	
	function deleteRedoStack(currentID)
	{
		if(Link.AnyExist("ScriptParams", currentID))
		{
			local spID = Link.GetOne("ScriptParams", currentID);
			local spLink = sLink(spID);
			local nextID = spLink.dest;
			deleteRedoStack(nextID);
		}
		Object.Destroy(currentID);
		Quest.Set("RedoObjID", 0);
	}
	
	function OnFrobWorldEnd()
	{
		if("ShowDropDown" in userparams())
		{
			showDropDown(userparams().ShowDropDown);
			return; //prevents showing a menu, which is a button, from adding an undo step
		}
		
		if("MenuSubProperty" in userparams())
		{
			if("Select" in userparams())
			{
				setListItem("SFXTest", userparams().MenuSubProperty, userparams().Select);
				deleteListItems(userparams().MenuSubProperty); //tidies them up after selection has been made
			}
			if("Cancel" in userparams())
			{
				deleteListItems(userparams().MenuSubProperty); //removes items and makes no changes
				updatePanelsWithVisualFeedback();
			}
			return; //prevents showing a menu, which is a button, from adding an undo step
		}
		
		updateUndoRedoStack();
		
		if("SubProperty" in userparams())
		{
			local particleProp = "ParticleGroup";
			local launchProp = "PGLaunchInfo";
			if("ChangeValue" in userparams())
			{
				switch(userparams().SubProperty)
				{		
					case "number of particles":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "size of particle":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "velocity":
						setVector(particleProp, userparams().SubProperty);
						break;
					case "gravity vector":
						setVector(particleProp,userparams().SubProperty);
						break;
					case "color (palettized)":
						setValuePositive(particleProp, userparams().SubProperty, 255);
						break;
					case "2nd color":
						setValuePositive(particleProp, userparams().SubProperty, 255);
						break;
					case "3rd color":
						setValuePositive(particleProp, userparams().SubProperty, 255);
						break;
					case "alpha":
						setValuePositive(particleProp, userparams().SubProperty, 255);
						break;
					case "fixed-group radius":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "spin speed":
						setVector(particleProp, userparams().SubProperty);
						break;
					case "pulse magnitude":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "pulse cycle time ms":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "particle fade time":
						setValueSFloat(particleProp, userparams().SubProperty);
						break;
					case "launch period":
						setValueSFloat(particleProp, userparams().SubProperty);
						break;
					case "animation offset ms":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "Group-scale velocity":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "bm-disk birth time":
						setValueSFloat(particleProp, userparams().SubProperty);
						break;
					case "bm-disk rot (vel, offs, rand)":
						setVector(particleProp, userparams().SubProperty);
						break;
					case "bm-disk ani frame time":
						setValueSFloat(particleProp, userparams().SubProperty);
						break;
					case "bm-disk grow speed":
						setValuePositive(particleProp, userparams().SubProperty);
						break;
					case "bm-disk rgb":
						setRGBColour(userparams().SubProperty);
						break;
					case "bm-disk 2nd rgb":
						setRGBColour(userparams().SubProperty);
						break;
					case "bm-disk 3rd rgb":
						setRGBColour(userparams().SubProperty);
						break;
					case "Box Min":
						setVector(launchProp, userparams().SubProperty);
						break;
					case "Box Max":
						setVector(launchProp, userparams().SubProperty);
						break;
					case "Min Radius":
						setValuePositive(launchProp, userparams().SubProperty);
						break;
					case "Max Radius":
						setValuePositive(launchProp, userparams().SubProperty);
						break;
					case "Velocity Min":
						setVector(launchProp, userparams().SubProperty);
						break;
					case "Velocity Max":
						setVector(launchProp, userparams().SubProperty);
						break;
					case "Min time":
						setValuePositive(launchProp, userparams().SubProperty);
						break;
					case "Max time":
						setValuePositive(launchProp, userparams().SubProperty);
						break;
				}
			}
			else if("ToggleValue" in userparams())
			{
				switch(userparams().SubProperty)
				{
					case "Active":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "spin / pulse group":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "always simulate (slow)":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "always simulate group":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "particles start launched":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "alpha subpixel particles":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "skip subpixel particles":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "ignore attachment refs":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "force matching unrefs":
						toggleBool("SFXTest", particleProp, userparams().SubProperty);
						break;
					case "Loc unrotated?":
						toggleBool("SFXTest", launchProp, userparams().SubProperty);
						break;
					case "Vel unrotated?":
						toggleBool("SFXTest", launchProp, userparams().SubProperty);
						break;
				}
			}
		}
		else if("Bitmaps" in userparams())
		{
			local bMapList = split(userparams().Bitmaps, ";");
			
			for(local i = 0; i < bMapList.len(); i++)
			{
				if(bMapList[i] == "null") bMapList[i] = ""; //allows the field to be cleared. User also just use an extra ; but null is easier to read
			}
			
			local currentItem = 0;
			local dataName = "currentBitmap";
			
			if(!IsDataSet(dataName)) SetData(dataName, currentItem);
			else currentItem = GetData(dataName);
			
			//increment or go back to the start
			if(currentItem < bMapList.len() -1) currentItem = GetData(dataName) + 1;
			else currentItem = 0;
			
			SetData(dataName, currentItem);

			Property.Set("SFXTest", "ParticleGroup", "bitmap name", bMapList[currentItem]);
		}
		else if("ToggleFlag" in userparams())
		{
			local flagValue = toggleBitmapDiskFlag(userparams().ToggleFlag);
			Property.Set("SFXTest", "ParticleGroup", "bm-disk flags", flagValue);
		}
		else if("Property" in userparams())
		{
			switch(userparams().Property)
			{
				case "RenderAlpha":
					local alpha = Property.Get("SFXTest", "RenderAlpha");
					local newAlpha = alpha += userparams().ChangeValue;
					if(newAlpha > 1) newAlpha = 1;
					else if(newAlpha < 0) newAlpha = 0;
					Property.SetSimple("SFXTest", "RenderAlpha", newAlpha);
					break;
			}
		}
		else if("DefaultProp" in userparams() && "DefaultSubProp" in userparams())
		{
			setArchetypeDefault(userparams().DefaultProp, userparams().DefaultSubProp);
		}
		else if("DefaultPropSimple" in userparams())
		{
			setArchetypeSimpleDefault(userparams().DefaultPropSimple);
		}
		
		updatePanelsWithVisualFeedback();
	}
	
	function updatePanelsWithVisualFeedback()
	{
		SendMessage("ParticlesPanel", "TurnOn"); //shows the current values of the particles property
		
		//To help confirm if the script is doing something without errors
		if(Object.Exists("LightRevert")) SendMessage("LightRevert", "TurnOn");
		Sound.PlaySchemaAtObject(self, "Button_push", self);
	}
	
	function setArchetypeSimpleDefault(prop)
	{
		local arc = Object.Archetype("SFXTest");
		local defaultPropValue = Property.Get (arc, prop);
		Property.SetSimple("SFXTest", prop, defaultPropValue);
	}
	
	function setArchetypeDefault(prop, subProp)
	{
		local arc = Object.Archetype("SFXTest");
		local defaultPropValue = Property.Get (arc, prop, subProp);
		
		if(subProp == "All") Property.CopyFrom("SFXTest", prop, arc);
		else Property.Set("SFXTest", prop, subProp, defaultPropValue);
	}
	
	function showDropDown(menuName)
	{
		Sound.PlaySchemaAtObject(self, "LvWall", self);
		local blockerObjectArchetype = "PanelBlockerPGroup"; //default, 75% of all dropdowns are for ParticleGroup
		local blockerConcreteName = "PGroupBlocker"; //default, 75% of all dropdowns are for ParticleGroup
		local blockerAcnhorObjName = "DropDownBlockPointPG"; //default, 75% of all dropdowns are for ParticleGroup
		local listMax = 0; 
		local itemNamePrefix = "";
		local cancelButton = ""
		
		switch(menuName)
		{
			case "Particle Render Type":
				listMax = 7;
				itemNamePrefix = "PRT_Item";
				cancelButton= "PRT_Cancel";
				break;
			case "Particle Animation":
				listMax = 9;
				itemNamePrefix = "PA_Item";
				cancelButton= "PA_Cancel";
				break;
			case "Particle Group Motion":
				listMax = 6;
				itemNamePrefix = "PGM_Item";
				cancelButton= "PGM_Cancel";
				break;
			case "Launch Type":
				blockerObjectArchetype = "PanelBlockerLaunchInfo";
				blockerConcreteName = "LInfoBlocker";
				blockerAcnhorObjName = "DropDownBlockPointLI";
				listMax = 3;
				itemNamePrefix = "LI_Item";
				cancelButton = "LI_Cancel";
				break;
		}
		
		//create blocker so it's easy to select menu items
		local blockerID = Object.BeginCreate(blockerObjectArchetype);
		Object.SetName(blockerID, blockerConcreteName);
		Object.Teleport(blockerID, vector(0,0,0), vector(0,0,0), blockerAcnhorObjName);
		Object.EndCreate(blockerID);
		
		//show menu items
		for(local i = 0; i < listMax; i++)
		{
			Property.SetSimple(itemNamePrefix + i, "RenderType", 0); //Normal
		}
		Property.SetSimple(cancelButton, "RenderType", 0); //Normal
	}
	
	function setRGBColour(subPropName)
	{	
		local colourProperty = Property.Get("SFXTest", "ParticleGroup", subPropName);
		local newColourArray = colPropToArray(colourProperty);

		switch(userparams().Component)
		{
			case "red":
				local newRed = newColourArray[0] + userparams().ChangeValue;
				newColourArray[0] = keepBetween0and255(newRed);
				break;
			case "green":
				local newGreen = newColourArray[1] + userparams().ChangeValue;
				newColourArray[1] = keepBetween0and255(newGreen);
				break;
			case "blue":
				local newBlue = newColourArray[2] + userparams().ChangeValue;
				newColourArray[2] = keepBetween0and255(newBlue);
				break;
		}
		
		local updatedArrayString = newColourArray[0] + ", " + newColourArray[1] + ", " + newColourArray[2];
		
		Property.Set("SFXTest", "ParticleGroup", subPropName, updatedArrayString);
	}
	
	function keepBetween0and255(testValue)
	{
		if(testValue > 255) return 255;
		else if(testValue < 0) return 0;
		else return testValue;
	}
	
	function toggleBitmapDiskFlag(flagNumber)
	{
		local currentFlagsValue = Property.Get("SFXTest", "ParticleGroup", "bm-disk flags");		
		if(currentFlagsValue & flagNumber) return currentFlagsValue -= flagNumber; //if flag is set, remove its value from the total
		else return currentFlagsValue += flagNumber; //if flag is not set, add its value to the total
	}
	
	function toggleBool(objName, propName, subPropName)
	{
		local chk = Property.Get(objName, propName, subPropName);
		if(chk) Property.Set(objName, propName, subPropName, false);
		else Property.Set(objName, propName, subPropName, true);
	}
	
	function setVector(objectPropName, subPropName)
	{
		local fullVector = Property.Get("SFXTest", objectPropName, subPropName);
		switch(userparams().Component)
		{
			case "x":
				fullVector.x = fullVector.x + userparams().ChangeValue;
				break;
			case "y":
				fullVector.y = fullVector.y + userparams().ChangeValue;
				break;
			case "z":
				fullVector.z = fullVector.z + userparams().ChangeValue;
				break;
		}
		Property.Set("SFXTest", objectPropName, subPropName, fullVector);
	}
	
	function setValuePositive(objectPropName, subPropToChange, maxValue = 0)
	{
		local currentValue = Property.Get("SFXTest", objectPropName, subPropToChange);
		local newValue = currentValue + userparams().ChangeValue;
		if(maxValue != 0) if(newValue > maxValue) newValue = maxValue;
		if(newValue < 0) newValue = 0; //prevents it going below 0
		Property.Set("SFXTest", objectPropName, subPropToChange, newValue);
	}
	
	function setValueSFloat(objectPropName, subPropToChange)
	{
		local currentValue = Property.Get("SFXTest", objectPropName, subPropToChange);
		local newValue = currentValue + userparams().ChangeValue * 65536;
		if(newValue < 0) newValue = 0; //prevents it going below 0
		Property.Set("SFXTest", objectPropName, subPropToChange, newValue);
	}
	
	function setListItem(targetObject, subPropName,  value)
	{
		updateUndoRedoStack();
		local propName = "ParticleGroup" //usal case for drop downs
		if(subPropName == "Launch Type") propName = "PGLaunchInfo";
		Property.Set(targetObject, propName, subPropName, value);
		
		updatePanelsWithVisualFeedback();
	}
	
	function deleteListItems(subPropName)
	{	
		local menuLimit = 0;
		local namePrefix = "";
		local blockerName = "";
		local cancelButton = "";
		
		switch(subPropName)
		{
			case "Particle Render Type":
				menuLimit = 7;
				namePrefix = "PRT_Item";
				blockerName = "PGroupBlocker";
				cancelButton = "PRT_Cancel";
				break;
			case "Particle Animation":
				menuLimit = 9;
				namePrefix = "PA_Item";
				blockerName = "PGroupBlocker";
				cancelButton = "PA_Cancel";
				break;
			case "Particle Group Motion":
				menuLimit = 6;
				namePrefix = "PGM_Item";
				blockerName = "PGroupBlocker";
				cancelButton = "PGM_Cancel";
				break;
			case "Launch Type":
				menuLimit = 3;
				namePrefix = "LI_Item";
				blockerName = "LInfoBlocker";
				cancelButton = "LI_Cancel";
				break;
		}
		
		Object.Destroy(blockerName);
		for(local i = 0; i < menuLimit; i++)
		{
			Property.SetSimple(namePrefix + i, "RenderType", 1); //not rendered
		}
		Property.SetSimple(cancelButton, "RenderType", 1); //not rendered
	}
}